/*
 * On macOS, build with:
 *   gcc -c helper_skeleton.c -DTARGET_OS_MAC -arch i386 -arch x86_64
 *   gcc -dynamiclib -undefined suppress -flat_namespace helper_skeleton.o -o C_Helper.dylib -arch i386 -arch x86_64
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <zlib.h>
#include <string.h>
#include <assert.h>

#if TARGET_OS_MAC
	#include <CoreFoundation/CoreFoundation.h>
	#define DLL
#else
	// Windows
	typedef signed char int8_t;
	typedef unsigned char uint8_t;
	typedef unsigned short uint16_t;
	typedef long int32_t;
	typedef unsigned long uint32_t;
	typedef unsigned long long uint64_t;
	#define DLL __declspec(dllexport) __stdcall
#endif

typedef int boolean;
#define false 0
#define true 1

static pthread_t *gThreads = NULL;
static int gThreadCount = 0;
static int gStop = 0;

typedef struct {	// links the sharedTask records into a queue
	uint16_t ident;				// must equal 0xB117
	uint16_t ownSize;			// must equal sizeof(privateData)
	struct sharedTask *next;
} privateData;

typedef struct sharedTask {	// for passing data between this code and Xojo code - requires a matching Structure in Xojo
	uint16_t ident;				// must equal 0xB119
	uint16_t ownSize;			// must equal sizeof(sharedTask)
	privateData internal;
	uint8_t status;				// 0: ready, 1: working, 2: finished, 3: error
	// ... add your own data here that you'd pass between Xojo and this code:
	char *inputBuf;
	uint32_t inputLen;
	uint64_t result;
} sharedTask;

static sharedTask *gNextTask = NULL;
static sharedTask *gLastTask = NULL;
static int gPendingTaskCount = 0;
static int gBusyTaskCount = 0;

static pthread_mutex_t gInputLock;
static pthread_cond_t gInputCond = PTHREAD_COND_INITIALIZER;

static void *threadRunner(void *threadNum)
{
	// This thread is kept alive as long as we might need it.
	// It will wait for new tasks in a loop, so that we do not
	// have to create new threads every time we want to have it
	// do some work.

	while (!gStop) {
		sharedTask *myTask;

		pthread_mutex_lock (&gInputLock);

		// Wait for a new task (added via helper_enqueue)
		while (gNextTask == NULL) {
			pthread_cond_wait (&gInputCond, &gInputLock);
			if (gStop) {
				pthread_mutex_unlock (&gInputLock);
				goto bail;
			}
		}

		// Pick next task from queue
		myTask = gNextTask;
		gNextTask = myTask->internal.next;
		gPendingTaskCount -= 1;

		pthread_mutex_unlock (&gInputLock);

		myTask->status = 1;
		gBusyTaskCount += 1;
		
		boolean ok = false;
		
		//
		// Do the task's work here, using the information in myTask
		//
		
		// As an example of a time-intensive task, let's count the
		// set bits in all the input buffer's bytes, with a rather
		// inefficient algorithm.
		uint64_t bitCount = 0; 
		for (int byteOfs = 0; byteOfs < myTask->inputLen; ++byteOfs) {
			for (int bit = 0; bit < 8; ++bit) {
				if ((myTask->inputBuf[byteOfs] & (1 << bit)) != 0) {
					bitCount += 1;
				}
			}
		}
		
		myTask->result = bitCount;
		ok = true;

		// Mark the task finished
		myTask->status = (ok) ? 2 : 3;
		gBusyTaskCount -= 1;
	}

bail:
	// cleanup...
	
	// wake other threads, so they check the gStop condition as well
	pthread_cond_signal (&gInputCond);

	return NULL;
}

DLL extern boolean helper_enqueue (sharedTask *task)
{
	assert (task->ident == 0xB119 && task->internal.ident == 0xB117);
	assert (task->ownSize == sizeof(*task));
	assert (task->internal.ownSize == sizeof(task->internal));

	task->internal.next = NULL;

	// Add the task to the queue.
	pthread_mutex_lock(&gInputLock);
	if (!gNextTask) {
		// The queue was empty - we start fresh
		gNextTask = task;
	} else {
		// Append this task to the end of the queue
		gLastTask->internal.next = task;
	}
	gLastTask = task;
	gPendingTaskCount += 1;
	pthread_mutex_unlock(&gInputLock);

	// Signal the waiting threads so that one picks up the task
	pthread_cond_signal (&gInputCond);

	return true;
}

DLL extern boolean helper_disposeThreads ()
{
	// Stop threads from processing more tasks
	pthread_mutex_lock (&gInputLock);
	gNextTask = NULL;
	gLastTask = NULL;
	gStop = 1;
	pthread_mutex_unlock (&gInputLock);

	if (gThreads) {
		// let's wake our threads so that they see the gStop flag and quit
		pthread_cond_signal (&gInputCond);

		// Wait for all threads to end before we free their memory
		for (int n = 0; n < gThreadCount; ++n) {
			pthread_t *th = &gThreads[n];
			pthread_join (*th, NULL);
		}

		free (gThreads);
	}
	gThreads = NULL;
	gThreadCount = 0;

	gPendingTaskCount = 0;
	gBusyTaskCount = 0;

	return true;
}

DLL extern boolean helper_createThreads (int threadCountIn)
{
	// The memory for the "pthread_t" elements needs to be static,
	// hence we allocate it now dynamically

	if (gThreads) {
		// Error: There's already one allocated
		return false;
	}

	gStop = 0;
	gThreads = calloc (threadCountIn, sizeof(gThreads[0]));
	gThreadCount = threadCountIn;

	for (int n = 0; n < gThreadCount; ++n) {
		pthread_create (&gThreads[n], NULL, threadRunner, (void*)(size_t)n);
	}

	return true;
}

DLL extern boolean helper_getCounts (int32_t *threadsOut, int32_t *pendingOut, int32_t *busyOut)
{
	*threadsOut = gThreadCount;
	*pendingOut = gPendingTaskCount;
	*busyOut = gBusyTaskCount;
	return true;
}

DLL extern boolean helper_initialize ()
{
	pthread_mutex_init (&gInputLock, NULL);
	return true;
}
